;+
; Procedure: tplot_force_monotonic
;
; Purpose:
;    This routine checks tplot variables for sample time (abscissa: data.x) monotonicity
;     and forces them so through removal or replacement of non-monotonic segments, if requested;
;     the corresponding elements of data.y and data.v (if 2d) are also repaired.
;    Indices of consecutively repeated header times and piece-wise monotonic segment "negative jumps"
;     are identified and used to determine monotonicity of tplot variables. If checking,
;     the routine will report cause of failures (i.e. negative jumps, repeats).  
;    Three repair methods are available: /forward, /reverse and /sort (see Keywords) 
;    
; Inputs:   [optional] tplot variable name(s) string/array or wild-card name string or tplot variable number(s); same as input for tnames()
;                    
; Keywords:
;           forward:      the recommended repair method, which keeps the older time elements of over-lapping time segments
;           reverse:      repair method which keeps the newer time elements over-lapping time segments
;           sort:         repair method which sorts time-lines chronologically, using bsort()
;           
;           keep_repeats: [optional] do not remove consecutively repeated header times
;           replace_NaN:  [optional] instead of removing 'bad' elements, replace them with NaN (only applied to data.y and data.v elements, data.x is not modified)
;           
;              
; Outputs:
;           If checking, PASS/FAIL console message per tplot variable, including cause of failures
;           If repairing, PASS/FAIL console message per tplot variable and repaired tplot variable(s) (via store_data) 
;              
; Examples:
;           tplot_force_monotonic                           ; check all tplot variables
;           tplot_force_monotonic,'*'                       ; check all tplot variables
;           tplot_force_monotonic,'var1'                    ; check tplot variable named var1
;           tplot_force_monotonic,['var1','var2']           ; check tplot variables named var1 and var2
;           tplot_force_monotonic,[6,7,9]                   ; check tplot variables 6, 7 and 9
;           tplot_force_monotonic,/forward                  ; repair all tplot variables using 'forward' method, discarding repeated sample times (recommended repair method)
;           tplot_force_monotonic,/reverse,/keep_repeats    ; repair all tplot variables using 'reverse' method, keeping repeated sample times
;           tplot_force_monotonic,/sort                     ; repair all tplot variables using 'sort' method, discarding consecutively repeated sample times
;           tplot_force_monotonic,'var?',/forward,/replace_nan  ; repair tplot variable(s) in var? using 'forward' method and replace bad elements with NaNs
;        
; Notes:
;     1. tplot variables repaired with the /replace_nan keyword will not pass a monotonicity check.
;     2. tplot variables repaired with the /keep_repeats keyword may not pass a monotonicity check.
;     3. A warning is issued if more than 10% of tplot variable elements are removed or replaced.
;      
; ToDo: nothing yet
;
; $LastChangedBy: $
; $LastChangedDate: $
; $LastChangedRevision: $
; $URL: $
;-

function keep_overlap, i_keep, i_new, time_array, REVERSE=REVERSE, KEEP_REPEATS=KEEP_REPEATS
;;; function which returns the indices of overlapping time segments to keep
if keyword_set(reverse) then begin
  if keyword_set(KEEP_REPEATS) then begin i_keep_new = where( time_array[i_new] le min(time_array[i_keep]) )
  endif else i_keep_new = where( time_array[i_new] lt min(time_array[i_keep]) )
endif else begin
  if keyword_set(KEEP_REPEATS) then begin i_keep_new = where( time_array[i_new] ge max(time_array[i_keep]) )
  endif else i_keep_new = where( time_array[i_new] gt max(time_array[i_keep]) )
endelse

if max(i_keep_new) ge 0 then i_keep = [i_keep, i_new[i_keep_new]]

return, i_keep
end

pro tplot_force_monotonic, tplot_vars, FORWARD=FORWARD, REVERSE=REVERSE, SORT=SORT, KEEP_REPEATS=KEEP_REPEATS, REPLACE_NAN=REPLACE_NAN

;;; Get and check tplot variable names/numbers
tplot_vars = tnames(tplot_vars)
if strlen(strjoin(tplot_vars)) eq 0 then begin
  message, 'Invalid tplot variable name(s) or number(s)',/continue
  return
endif
max_var_len = 32

;;; Loop over requested tplot variables
for j=0L,n_elements(tplot_vars)-1L do begin
  tplot_var = tplot_vars[j]
  get_data,tplot_var,data=data,dlimits=dlimits,limits=limits
  time_array = data.x

  ;;; Check if time array has only one element
  if (n_elements(time_array) eq 1) then begin
    dprint,' ',tplot_var,':'+strjoin(strarr(max_var_len-strlen(tplot_var))+' ')+'PASS, variable has one time element'
    continue
  endif
  
  ;;; Identify consecutive repeats and boundaries of piece-wise monotonic segments "negative jumps"
  i_repeat = where( (time_array - shift(time_array,1)) eq 0d)
  i_non_monotonic = where( (time_array - shift(time_array,1)) lt 0d)

  ;;; If time_array is monotonic (i.e. has no negative jumps) and has no repeats, then it PASSES
  if (n_elements(i_non_monotonic) eq 1) and (min(i_repeat) eq -1) then begin 
    dprint,' ',tplot_var,':'+strjoin(strarr(max_var_len-strlen(tplot_var))+' ')+'PASS'
    continue
  endif else begin
    
    ;;; case: tplot variable is non-monotonic
    if keyword_set(forward) or keyword_set(reverse) or keyword_set(SORT) then begin
      if n_elements(i_non_monotonic) eq 1 then begin
        GOTO, ONLY_REPEATS
      endif else i_non_monotonic = i_non_monotonic[1:*]
      
      ;;; Keep forward portion of overlaps
      if keyword_set(forward) then begin
        i_keep = lindgen(i_non_monotonic[0])
        for k = 0L, n_elements(i_non_monotonic)-2L do begin
            i_new = lindgen(i_non_monotonic[k+1] - i_non_monotonic[k]) + i_non_monotonic[k]
            i_keep = keep_overlap(i_keep,i_new,time_array,KEEP_REPEATS=KEEP_REPEATS)
        endfor
        i_new = lindgen( n_elements(time_array) - i_non_monotonic[k] ) + i_non_monotonic[k]
        i_keep = keep_overlap(i_keep,i_new,time_array,KEEP_REPEATS=KEEP_REPEATS)
      endif 
    
      ;;; Keep reverse portion of overlaps
      if keyword_set(reverse) then begin
        n = n_elements(i_non_monotonic)-1
        i_keep = lindgen( n_elements(time_array) - i_non_monotonic[n]) +i_non_monotonic[n]
        for k = n_elements(i_non_monotonic)-2L,0,-1 do begin
            i_new = lindgen(i_non_monotonic[k+1]-i_non_monotonic[k]) + i_non_monotonic[k]
            i_keep = keep_overlap(i_keep,i_new,time_array,/reverse,KEEP_REPEATS=KEEP_REPEATS)
        endfor
        i_new = lindgen(i_non_monotonic[0])
        i_keep = keep_overlap(i_keep,i_new,time_array,/reverse,KEEP_REPEATS=KEEP_REPEATS)
      endif
      
      ;;; Sort time-series using bsort()
      if keyword_set(SORT) then i_keep = bsort(time_array)
        
      ;;; Keep repeats or not
      ONLY_REPEATS: if min(i_non_monotonic) le 0 then i_keep = lindgen(n_elements(time_array))
      if not keyword_set(KEEP_REPEATS) then begin
        i_non_repeats = ssl_set_complement([i_repeat],[i_keep])
        if min(i_non_repeats) ne -1 then i_keep = i_non_repeats
      endif
      
      ;;; Repair tplot variable and print summary
        ;;; Remove "bad" elements of data.x, data.y and data.v
      if not keyword_set(replace_nan) then begin
        x = data.x[i_keep]
        y_dim = size(data.y,/n_dimensions)
        if y_dim eq 1 then begin
          y = data.y[i_keep]
          if tag_exist(data,'v') then v = data.v[i_keep]
        endif
        if y_dim eq 2 then begin
          y = data.y[i_keep,*]
          if tag_exist(data,'v') then begin
            v_dim = size(data.v,/n_dimensions)
            if v_dim eq 2 then begin
              v = data.v[i_keep,*]
            endif else v = data.v
          endif
        endif
        if y_dim eq 3 then begin
          y = data.y[i_keep,*,*]
          if tag_exist(data,'v1') then begin
            v1 = data.v1
            v2 = data.v2
          endif
        endif
        ;;; Replace "bad" elements of data.y and data.v with NaN
      endif else begin
        x = data.x
        i_all = lindgen(n_elements(x))
        i_replace = ssl_set_complement([i_keep],[i_all])
        if min(i_replace) ne -1 then begin
          y_dim = size(data.y,/n_dimensions)
          if y_dim eq 1 then begin
            y = data.y
            y[i_replace] = 'NaN'
            if tag_exist(data,'v') then begin
              v = data.v
              v[i_replace] = 'NaN'
            endif
          endif
          if y_dim eq 2 then begin
            y = data.y
            y[i_replace,*] = 'NaN'
            if tag_exist(data,'v') then begin
              v = data.v
              v_dim = size(data.v,/n_dimensions)
              if v_dim eq 2 then v[i_replace,*] = 'NaN'
            endif
          endif
          if y_dim eq 3 then begin
            y = data.y
            y[i_replace,*,*] = 'NaN'
            if tag_exist(data,'v1') then begin
              v1 = data.v1
              v2 = data.v2
            endif
          endif
        endif            
       endelse
      
      ;;; Write modified data to tplot variable
      if (tag_exist(data,'v') or tag_exist(data,'v1')) then begin
        if tag_exist(data,'v') then store_data,tplot_var,data={x:x,y:y,v:v},dlimits=dlimits,limits=limits
        if tag_exist(data,'v1') then store_data,tplot_var,data={x:x,y:y,v1:v1,v2:v2},dlimits=dlimits,limits=limits
      endif else store_data,tplot_var,data={x:x,y:y},dlimits=dlimits,limits=limits
      
      ;;; Determine percentage of elements removed or replaced, and report them
      percent_removed = (n_elements(time_array) - n_elements(i_keep)) / float(n_elements(time_array)) * float(100)
      if not keyword_set(replace_nan) then begin
        dprint,tplot_var,':'+strjoin(strarr(max_var_len-strlen(tplot_var))+' ')+'FAIL - repaired: '+ strtrim(percent_removed,2) + ' % removed'
      endif else dprint,tplot_var,':'+strjoin(strarr(max_var_len-strlen(tplot_var))+' ')+'FAIL - repaired: '+ strtrim(percent_removed,2) + ' % replaced with NaN'   
      if percent_removed gt 10d then message,'********** Warning: Over 10% removed or replaced from '+strtrim(tplot_var,2)+' **********',/continue
    endif else begin
      ;;; Checking case: report causes of failures
      if ((min(i_repeat) ne -1) and (n_elements(i_non_monotonic) ne 1)) then dprint,tplot_var,':'+strjoin(strarr(max_var_len-strlen(tplot_var))+' ')+'FAIL - negative jumps and repeats'
      if ((min(i_repeat) ne -1) and not (n_elements(i_non_monotonic) ne 1)) then dprint,tplot_var,':'+strjoin(strarr(max_var_len-strlen(tplot_var))+' ')+'FAIL - repeats'
      if ( not (min(i_repeat) ne -1) and (n_elements(i_non_monotonic) ne 1)) then dprint,tplot_var,':'+strjoin(strarr(max_var_len-strlen(tplot_var))+' ')+'FAIL - negative jumps'
    endelse
  endelse

endfor  ;;; end of loop over tplot variables
end